@@ -1,3 +1,5 @@
+//! Cooperative coroutines in task.
+
 mod context;
 mod page_size;
 pub(crate) mod stack;
@@ -8,12 +10,13 @@ use std::panic::{self, AssertUnwindSafe};
 use std::rc::Rc;
 use std::{mem, ptr};
 
+use ignore_result::Ignore;
 use static_assertions::assert_not_impl_any;
 
 use self::context::{Context, Entry};
 use self::stack::StackSize;
 use crate::error::{JoinError, PanicError};
-use crate::task;
+use crate::task::{self, Yielding};
 
 thread_local! {
     static COROUTINE: Cell<Option<ptr::NonNull<Coroutine>>> = Cell::new(None);
@@ -74,7 +77,6 @@ impl ThisThread {
 pub(crate) struct Coroutine {
     context: Box<Context>,
     completed: bool,
-    panicking: Option<&'static str>,
     f: Option<Box<dyn FnOnce()>>,
 }
 
@@ -87,7 +89,6 @@ impl Coroutine {
             f: Option::Some(f),
             context: unsafe { mem::MaybeUninit::zeroed().assume_init() },
             completed: false,
-            panicking: None,
         });
         let entry = Entry { f: Self::main, arg: (co.as_mut() as *mut Coroutine) as *mut libc::c_void, stack_size };
         mem::forget(mem::replace(&mut co.context, Context::new(&entry, None)));
@@ -106,10 +107,6 @@ impl Coroutine {
         f();
     }
 
-    pub fn set_panic(&mut self, msg: &'static str) {
-        self.panicking = Some(msg);
-    }
-
     /// Resumes coroutine.
     ///
     /// Returns whether this coroutine should be resumed again.
@@ -121,58 +118,102 @@ impl Coroutine {
 
     pub fn suspend(&mut self) {
         ThisThread::suspend(&mut self.context);
-        if let Some(msg) = self.panicking {
-            panic::panic_any(msg);
-        }
     }
 }
 
 enum SuspensionState<T: 'static> {
     Empty,
     Value(T),
-    Panicked(Box<dyn Any + Send + 'static>),
+    Panicked(PanicError),
     Joining(ptr::NonNull<Coroutine>),
+    Joined,
 }
 
 struct SuspensionJoint<T: 'static> {
     state: UnsafeCell<SuspensionState<T>>,
+    wakers: Cell<usize>,
+}
+
+impl<T> Yielding for SuspensionJoint<T> {
+    fn interrupt(&self, reason: &'static str) -> bool {
+        self.cancel(PanicError::Static(reason));
+        true
+    }
 }
 
 impl<T> SuspensionJoint<T> {
     fn new() -> Rc<SuspensionJoint<T>> {
-        Rc::new(SuspensionJoint { state: UnsafeCell::new(SuspensionState::Empty) })
+        Rc::new(SuspensionJoint { state: UnsafeCell::new(SuspensionState::Empty), wakers: Cell::new(1) })
     }
 
-    fn resume(&self, state: SuspensionState<T>) {
-        let state = mem::replace(unsafe { &mut *self.state.get() }, state);
-        if let SuspensionState::Joining(co) = state {
-            let task = unsafe { task::current().as_mut() };
-            task.resume(co);
+    fn is_ready(&self) -> bool {
+        let state = unsafe { &*self.state.get() };
+        matches!(state, SuspensionState::Value(_) | SuspensionState::Panicked(_) | SuspensionState::Joined)
+    }
+
+    fn wake_coroutine(co: ptr::NonNull<Coroutine>) {
+        let task = unsafe { task::current().as_mut() };
+        task.resume(co);
+    }
+
+    fn add_waker(&self) {
+        let wakers = self.wakers.get() + 1;
+        self.wakers.set(wakers);
+    }
+
+    fn remove_waker(&self) {
+        let wakers = self.wakers.get() - 1;
+        self.wakers.set(wakers);
+        if wakers == 0 {
+            self.fault(PanicError::Static("suspend: no resumption"));
         }
     }
 
-    fn abort(&self) {
-        self.resume(SuspensionState::Panicked(Box::new("suspend: no resumption")))
+    fn cancel(&self, err: PanicError) -> Option<ptr::NonNull<Coroutine>> {
+        let state = unsafe { &mut *self.state.get() };
+        if matches!(state, SuspensionState::Value(_) | SuspensionState::Panicked(_) | SuspensionState::Joined) {
+            return None;
+        }
+        let state = unsafe { ptr::replace(state, SuspensionState::Panicked(err)) };
+        if let SuspensionState::Joining(co) = state {
+            return Some(co);
+        }
+        None
     }
 
-    fn wake(&self, value: T) {
-        self.resume(SuspensionState::Value(value));
+    fn fault(&self, err: PanicError) {
+        if let Some(co) = self.cancel(err) {
+            Self::wake_coroutine(co);
+        }
+    }
+
+    pub fn wake(&self, value: T) -> Result<(), T> {
+        let state = unsafe { &mut *self.state.get() };
+        if matches!(state, SuspensionState::Value(_) | SuspensionState::Panicked(_) | SuspensionState::Joined) {
+            return Err(value);
+        }
+        let state = unsafe { ptr::replace(state, SuspensionState::Value(value)) };
+        if let SuspensionState::Joining(co) = state {
+            Self::wake_coroutine(co);
+        }
+        Ok(())
     }
 
     fn set_result(&self, result: Result<T, Box<dyn Any + Send + 'static>>) {
-        let state = match result {
-            Ok(value) => SuspensionState::Value(value),
-            Err(err) => SuspensionState::Panicked(err),
-        };
-        self.resume(state);
+        match result {
+            Ok(value) => self.wake(value).ignore(),
+            Err(err) => self.fault(PanicError::Unwind(err)),
+        }
     }
 
     fn take(&self) -> Result<T, PanicError> {
-        let state = mem::replace(unsafe { &mut *self.state.get() }, SuspensionState::Empty);
+        let state = mem::replace(unsafe { &mut *self.state.get() }, SuspensionState::Joined);
         match state {
             SuspensionState::Value(value) => Ok(value),
             SuspensionState::Panicked(err) => Err(err),
-            _ => unreachable!("suspend: unexpected branch"),
+            SuspensionState::Empty => unreachable!("suspension: empty state"),
+            SuspensionState::Joining(_) => unreachable!("suspension: joining state"),
+            SuspensionState::Joined => unreachable!("suspension: joined state"),
         }
     }
 
@@ -182,12 +223,13 @@ impl<T> SuspensionJoint<T> {
         match state {
             SuspensionState::Empty => {
                 let task = unsafe { task::current().as_mut() };
-                task.suspend(co);
+                task.suspend(co, self);
                 self.take()
             },
             SuspensionState::Value(value) => Ok(value),
             SuspensionState::Panicked(err) => Err(err),
-            _ => unreachable!("suspend: unexpected branch"),
+            SuspensionState::Joining(_) => unreachable!("suspension: join joining state"),
+            SuspensionState::Joined => unreachable!("suspension: join joined state"),
         }
     }
 }
@@ -198,47 +240,90 @@ pub struct Suspension<T: 'static>(Rc<SuspensionJoint<T>>);
 /// Resumption provides method to resume suspending coroutine.
 pub struct Resumption<T: 'static> {
     joint: Rc<SuspensionJoint<T>>,
-    resumed: bool,
 }
 
 assert_not_impl_any!(Suspension<()>: Send);
 assert_not_impl_any!(Resumption<()>: Send);
 
 impl<T> Suspension<T> {
+    unsafe fn into_joint(self) -> Rc<SuspensionJoint<T>> {
+        let joint = ptr::read(&self.0);
+        mem::forget(self);
+        joint
+    }
+
+    /// Checks readiness.
+    pub fn is_ready(&self) -> bool {
+        self.0.is_ready()
+    }
+
     /// Suspends calling coroutine until [Resumption::resume].
     ///
     /// # Panics
     /// Panic if no resume from [Resumption].
+    ///
+    /// # Guarantee
+    /// Only two situations can happen:
+    /// * This method panics and no value sent
+    /// * This method returns and only one value sent
+    ///
+    /// This means that no value linger after panic.
     pub fn suspend(self) -> T {
-        match self.0.join() {
+        let joint = unsafe { self.into_joint() };
+        match joint.join() {
             Ok(value) => value,
-            Err(err) => panic::resume_unwind(err),
+            Err(PanicError::Unwind(err)) => panic::resume_unwind(err),
+            Err(PanicError::Static(s)) => panic::panic_any(s),
         }
     }
 }
 
+impl<T> Drop for Suspension<T> {
+    fn drop(&mut self) {
+        self.0.cancel(PanicError::Static("suspension dropped"));
+    }
+}
+
 impl<T> Resumption<T> {
     fn new(joint: Rc<SuspensionJoint<T>>) -> Self {
-        Resumption { joint, resumed: false }
+        Resumption { joint }
+    }
+
+    // SAFETY: Forget self and cancel drop to wake peer with no suspension interleave.
+    unsafe fn into_joint(self) -> Rc<SuspensionJoint<T>> {
+        let joint = ptr::read(&self.joint);
+        mem::forget(self);
+        joint
     }
 
     /// Resumes suspending coroutine.
-    pub fn resume(mut self, value: T) {
-        self.resumed = true;
-        self.joint.wake(value);
+    pub fn resume(self, value: T) -> bool {
+        let joint = unsafe { self.into_joint() };
+        joint.wake(value).is_ok()
     }
 
-    fn set_result(mut self, result: Result<T, Box<dyn Any + Send + 'static>>) {
-        self.resumed = true;
-        self.joint.set_result(result);
+    /// Sends and wakes peer if not waked.
+    pub fn send(self, value: T) -> Result<(), T> {
+        let joint = unsafe { self.into_joint() };
+        joint.wake(value)
+    }
+
+    fn set_result(self, result: Result<T, Box<dyn Any + Send + 'static>>) {
+        let joint = unsafe { self.into_joint() };
+        joint.set_result(result);
+    }
+}
+
+impl<T> Clone for Resumption<T> {
+    fn clone(&self) -> Self {
+        self.joint.add_waker();
+        Resumption { joint: self.joint.clone() }
     }
 }
 
 impl<T> Drop for Resumption<T> {
     fn drop(&mut self) {
-        if !self.resumed {
-            self.joint.abort()
-        }
+        self.joint.remove_waker();
     }
 }
 
@@ -251,14 +336,15 @@ pub fn suspension<T>() -> (Suspension<T>, Resumption<T>) {
 
 /// JoinHandle provides method to retrieve result of associated cooperative task.
 pub struct JoinHandle<T: 'static> {
-    joint: Rc<SuspensionJoint<T>>,
+    suspension: Suspension<T>,
 }
 
 assert_not_impl_any!(JoinHandle<()>: Send);
 
 impl<T> JoinHandle<T> {
     pub fn join(self) -> Result<T, JoinError> {
-        self.joint.join().map_err(|err| JoinError::new(err))
+        let joint = unsafe { self.suspension.into_joint() };
+        joint.join().map_err(JoinError::new)
     }
 }
 
@@ -270,10 +356,8 @@ where
     T: 'static,
 {
     let mut task = task::current();
-    let joint = SuspensionJoint::new();
-    let handle = JoinHandle { joint: joint.clone() };
-    // Wrap joint to abort JoinHandle::join on Resumption::drop.
-    let resumption = Resumption::new(joint);
+    let (suspension, resumption) = suspension();
+    let handle = JoinHandle { suspension };
     let f = Box::new(move || {
         let result = panic::catch_unwind(AssertUnwindSafe(f));
         resumption.set_result(result);
@@ -297,13 +381,11 @@ mod tests {
 
     use pretty_assertions::assert_eq;
 
-    use crate::coroutine;
-    use crate::runtime::Runtime;
+    use crate::{coroutine, task};
 
-    #[test]
+    #[crate::test(crate = "crate")]
     fn yield_now() {
-        let runtime = Runtime::new();
-        let five = runtime.spawn(|| {
+        let five = task::spawn(|| {
             let value = Rc::new(Cell::new(0));
             let shared_value = value.clone();
             coroutine::spawn(move || {
@@ -314,4 +396,40 @@ mod tests {
         });
         assert_eq!(5, five.join().unwrap());
     }
+
+    #[crate::test(crate = "crate")]
+    fn resumption() {
+        let (suspension, resumption) = coroutine::suspension();
+        drop(resumption.clone());
+        assert_eq!(suspension.0.is_ready(), false);
+        let co1 = coroutine::spawn({
+            let resumption = resumption.clone();
+            move || resumption.send(5)
+        });
+        let co2 = coroutine::spawn(move || resumption.send(6));
+        let value = suspension.suspend();
+        let mut result1 = co1.join().unwrap();
+        let mut result2 = co2.join().unwrap();
+        if result1.is_err() {
+            std::mem::swap(&mut result1, &mut result2);
+        }
+        assert_eq!(result1, Ok(()));
+        assert_eq!(result2.is_err(), true);
+        assert_eq!(value, 11 - result2.unwrap_err());
+    }
+
+    #[crate::test(crate = "crate")]
+    fn suspension_dropped() {
+        let (suspension, resumption) = coroutine::suspension::<()>();
+        drop(suspension);
+        assert_eq!(resumption.joint.is_ready(), true);
+    }
+
+    #[crate::test(crate = "crate")]
+    fn panic() {
+        const REASON: &'static str = "oooooops";
+        let co = coroutine::spawn(|| panic!("{}", REASON));
+        let err = co.join().unwrap_err();
+        assert!(err.to_string().contains(REASON))
+    }
 }
