@@ -8,9 +8,10 @@ use std::{ptr, thread};
 
 use ignore_result::Ignore;
 
-use crate::task::mpsc::{self, Sender};
+use crate::channel::parallel::{self, Sender};
+use crate::channel::prelude::*;
 use crate::task::{self, SchedFlow, Task};
-use crate::time;
+use crate::{net, time};
 
 thread_local! {
     static SCHEDULER: Cell<Option<ptr::NonNull<Scheduler>>> = Cell::new(None);
@@ -67,17 +68,25 @@ impl Builder {
         let parallelism = self
             .parallelism
             .unwrap_or_else(|| thread::available_parallelism().unwrap_or(NonZeroUsize::new(4).unwrap()).get());
-        let (time_sender, time_receiver) = mpsc::unbounded(512);
-        let scheduler = Scheduler::new(parallelism, time_sender.clone());
+        let (time_sender, time_receiver) = parallel::unbounded(512);
+        let poller = net::Poller::new().unwrap();
+        let scheduler = Scheduler::new(parallelism, time_sender.clone(), poller.registry());
+        let stopper = poller.start().unwrap();
         let timer = time::Timer::new();
         let timer = task::Builder::with_scheduler(&scheduler).spawn(move || {
             time::timer(timer, time_receiver);
         });
         let ticker = thread::spawn(move || {
-            time::tick(time_sender);
+            time::tickr(time_sender);
         });
         let scheduling_threads = Scheduler::start(&scheduler);
-        Runtime { scheduler, timer: MaybeUninit::new(timer), ticker: MaybeUninit::new(ticker), scheduling_threads }
+        Runtime {
+            scheduler,
+            timer: MaybeUninit::new(timer),
+            ticker: MaybeUninit::new(ticker),
+            stopper: MaybeUninit::new(stopper),
+            scheduling_threads,
+        }
     }
 }
 
@@ -88,6 +97,7 @@ pub struct Runtime {
     scheduler: Arc<Scheduler>,
     timer: MaybeUninit<task::JoinHandle<()>>,
     ticker: MaybeUninit<thread::JoinHandle<()>>,
+    stopper: MaybeUninit<net::Stopper>,
     scheduling_threads: Vec<thread::JoinHandle<()>>,
 }
 
@@ -105,7 +115,7 @@ impl Runtime {
     /// Spawns a concurrent task and returns a [task::JoinHandle] for it.
     ///
     /// See [task::spawn] for more details
-    pub fn spawn<F, T>(&self, f: F) -> task::JoinHandle<T>
+    pub fn spawn<F, T>(&mut self, f: F) -> task::JoinHandle<T>
     where
         F: FnOnce() -> T,
         F: Send + 'static,
@@ -126,8 +136,10 @@ impl Drop for Runtime {
         self.scheduler.stop();
         let timer = unsafe { ptr::read(self.timer.as_ptr()) };
         let ticker = unsafe { ptr::read(self.ticker.as_ptr()) };
+        let mut stopper = unsafe { ptr::read(self.stopper.as_ptr()) };
         timer.join().ignore();
         ticker.join().ignore();
+        stopper.stop();
         self.scheduler.stop();
         for handle in self.scheduling_threads.drain(..) {
             handle.join().ignore();
@@ -156,14 +168,21 @@ pub(crate) struct Scheduler {
     timer: Sender<time::Message>,
     state: Mutex<SchedulerState>,
     waker: Condvar,
+    registry: Arc<net::Registry>,
 }
 
 unsafe impl Send for Scheduler {}
 unsafe impl Sync for Scheduler {}
 
 impl Scheduler {
-    fn new(parallelism: usize, timer: Sender<time::Message>) -> Arc<Scheduler> {
-        Arc::new(Scheduler { parallelism, timer, state: Mutex::new(SchedulerState::new()), waker: Condvar::new() })
+    fn new(parallelism: usize, timer: Sender<time::Message>, registry: Arc<net::Registry>) -> Arc<Scheduler> {
+        Arc::new(Scheduler {
+            parallelism,
+            timer,
+            state: Mutex::new(SchedulerState::new()),
+            waker: Condvar::new(),
+            registry,
+        })
     }
 
     /// Starts threads to serve spawned tasks.
@@ -185,6 +204,10 @@ impl Scheduler {
         self.waker.notify_all();
     }
 
+    pub unsafe fn registry<'a>() -> &'a net::Registry {
+        &Self::current().registry
+    }
+
     pub(crate) unsafe fn current<'a>() -> &'a Scheduler {
         SCHEDULER.with(|s| s.get().unwrap_unchecked().as_ref())
     }
@@ -250,6 +273,7 @@ impl Scheduler {
         while state.stopped == self.parallelism as isize {
             state = self.run(state)
         }
+        // No timer and io poller now, this is the sole execution thread.
         while !state.registry.is_empty() {
             // SAFETY: Avoid compilation warning in read to `registry` and write to `runq`.
             let registry: &HashMap<u64, Arc<Task>> = unsafe { std::mem::transmute::<_, _>(&state.registry) };
@@ -288,7 +312,7 @@ mod tests {
 
     #[test]
     fn runtime_builder_parallelism_one() {
-        let runtime = Builder::default().parallelism(1).build();
+        let mut runtime = Builder::default().parallelism(1).build();
         let secret = 333;
         let set_secret = runtime.spawn(move || {
             thread::sleep(Duration::from_secs(10));
@@ -301,7 +325,7 @@ mod tests {
 
     #[test]
     fn runtime_builder_parallelism_multiple() {
-        let runtime = Builder::default().parallelism(2).build();
+        let mut runtime = Builder::default().parallelism(2).build();
         let secret = 111;
         let set_secret = runtime.spawn(move || {
             thread::sleep(Duration::from_secs(10));
